#
# Basis Encoding for full DATA and MOTI AGCT
#
#
# called from external: quantum_ram_2qubits_basis_encode for SNPs
#
#
#


from qiskit.circuit.library import RYGate
from qiskit.visualization import circuit_drawer

from qiskit import QuantumCircuit, transpile, assemble, QuantumRegister
from qiskit_aer import Aer, AerSimulator
from qiskit.visualization import plot_histogram
import matplotlib.pyplot as plt


import matplotlib.pyplot  as plt
import numpy as np
import math
import csv

from util import logging_utils

logger = logging_utils.setup_logger('qc_logger', 'qc_log.log')


def index_to_qubit_states(index, num_qubits):
    binary_index = bin(index)[2:].zfill(num_qubits)
    return [int(bit) for bit in binary_index]

def flip_qubits(qc, qubits_states):
    for i, state in enumerate(qubits_states):
        if state == 1:
            qc.x(i)

def apply_controlled_not_gate(qc, control_qubits, target_qubit):
    qc.mcx(control_qubits, target_qubit)


def qcm_basis_basis(qc, data1A, data1B, data1C, data1D, data2A, data2B, data2C, data2D, num_qubits):

    state_vector_size = 2 ** num_qubits

    # Create superposition (Hadamard gates on addressing qubits)
    for i in range(num_qubits - 8):  # Adjust according to 8 data qubits
        qc.h(i)



    # Iterate over each data index
    for i in range(len(data1A)):
        qubits_index = index_to_qubit_states(i, num_qubits - 8)

        # Check all 8 data inputs at index i
        if ('1' in [data1A[i], data1B[i], data1C[i], data1D[i], 
                    data2A[i], data2B[i], data2C[i], data2D[i]]):

            print(f"Data Index {i}: data1A={data1A[i]}, data1B={data1B[i]}, data1C={data1C[i]}, data1D={data1D[i]}, data2A={data2A[i]}, data2B={data2B[i]}, data2C={data2C[i]}, data2D={data2D[i]}")
            
            flip_qubits(qc, qubits_index)

            # Apply Controlled-NOT gates for each bit that is '1'
            data_bits = [data1A[i], data1B[i], data1C[i], data1D[i], data2A[i], data2B[i], data2C[i], data2D[i]]
            for bit_idx, bit in enumerate(data_bits):
                if bit == '1':
                    # Target qubit indices (last 8 qubits are data)
                    target_qubit = (num_qubits - 8) + bit_idx
                    apply_controlled_not_gate(qc, list(range(num_qubits - 8)), target_qubit)
            
            flip_qubits(qc, qubits_index)
            qc.barrier()


    return qc





def quantum_ram_basis_and_basis_8data (data1A, data1B, data1C, data1D, data2A, data2B, data2C, data2D):
    # Calculate the number of qubits needed for addressing based on the length of the encoding data
    # Qaddr = log base 2 N
    num_address_qubits = int(math.ceil(math.log(len(data1A), 2))) 

    # 2 qubits for (data1 and data2)
    total_qubits = num_address_qubits + 8

    print (f"data len = {len(data1A)}, total address qubit = {num_address_qubits}")

    # Define quantum registers explicitly
    address_reg = QuantumRegister(num_address_qubits, name='addr')
    data_reg = QuantumRegister(8, name='data')  # data1A to data2D represented as 8 qubits

    # Create a quantum circuit with the defined registers
    qc = QuantumCircuit(address_reg, data_reg)

    #print(qc)


    qcm_basis_basis (qc,data1A, data1B, data1C, data1D, data2A, data2B, data2C, data2D, total_qubits)

    return qc, total_qubits

