# main.py

from util import logging_utils
from hybrid_encoding import quantum_ram_basis_and_basis_8data
from compare import quantum_compare
from qiskit import QuantumCircuit, transpile, assemble
from qiskit_aer import Aer, AerSimulator
from qiskit.visualization import plot_histogram
import matplotlib.pyplot as plt

import time
import datetime
import numpy as np
import matplotlib.pyplot  as plt
import math
import csv

logger = logging_utils.setup_logger('qc_logger', 'qc_log.log')



def main():

    platform = 1

    if platform == 0: #windows
        data_path = "D:\source\KMUTNB\Qnx\data\data.dat"
       
    else: #linux
        data_path = "/Users/we/code/qlab/lab1/src/hybrid/data.dat"



def load_and_check_data(file_path):
    try:
        with open(file_path, "r") as file:
            lines = file.readlines()
    except Exception as e:
        print(f"Error loading data file: {e}")
        return (None,) * 8  # Returns eight None values

    expected_labels = ["Data1A", "Data1B", "Data1C", "Data1D",
                       "Data2A", "Data2B", "Data2C", "Data2D"]
    
    if len(lines) < len(expected_labels):
        print(f"Invalid data format. Expecting exactly {len(expected_labels)} lines.")
        return (None,) * 8

    data_values = []
    for i, label in enumerate(expected_labels):
        if '=' not in lines[i]:
            print(f"Invalid format in line {i+1}. Expected '{label}=...'")
            return (None,) * 8
        line_label, value = lines[i].strip().split('=', 1)
        if line_label != label:
            print(f"Unexpected label '{line_label}' at line {i+1}, expected '{label}'.")
            return (None,) * 8
        if not set(value).issubset({'0', '1'}):
            print(f"Invalid binary string for '{label}' at line {i+1}.")
            return (None,) * 8
        data_values.append(value)

    # Print sizes for verification
    for label, data in zip(expected_labels, data_values):
        print(f"Size of {label}: {len(data)}")

    return tuple(data_values)

def main():
    data_path = "data.dat"
    data1A, data1B, data1C, data1D, data2A, data2B, data2C, data2D = load_and_check_data(data_path)

    if None not in [data1A, data1B, data1C, data1D, data2A, data2B, data2C, data2D]:
        print("Data successfully loaded and validated.")
        print(f"data1A: {data1A}")
        print(f"data1B: {data1B}")
        print(f"data1C: {data1C}")
        print(f"data1D: {data1D}")
        print(f"data2A: {data2A}")
        print(f"data2B: {data2B}")
        print(f"data2C: {data2C}")
        print(f"data2D: {data2D}")
    else:
        print("Failed to load and validate data correctly.")


    qc, total = quantum_ram_basis_and_basis_8data(data1A, data1B, data1C, data1D, data2A, data2B, data2C, data2D)

    quantum_compare(qc, total-1, total-2, total-3, total-4, total-5, total-6, total-7, total-8) 

    qc.measure_all()

    print(qc)

    #For execution
    simulator = AerSimulator()
    #compiled_circuit = transpile(qc, simulator)
    sim_result = simulator.run(qc).result()
    counts = sim_result.get_counts()

    print("Result:", counts)

    #print (compiled_circuit)
    #plot_histogram(counts)
    #plt.show()

    interpret_results(qc, sim_result, len(data1A))


def interpret_results(qc, results, data_length):
   
    total_qubits = qc.num_qubits
    num_address_qubits = int(np.ceil(np.log2(data_length)))
    num_data_qubits = 8

    histogram_data = results.get_counts(qc)


    decoded_results = []

    for state, count in histogram_data.items():
        databits = state[:num_data_qubits]  # First 2 bits are data bits
        address_bits = state[num_data_qubits:num_data_qubits + num_address_qubits]  # remaining bits are address bits

        # Flip and reverse address bits to match the original indexing
        flipped_address = ''.join('1' if bit == '0' else '0' for bit in address_bits)
        reversed_address = flipped_address[::-1]

        decimal_address = int(reversed_address, 2)

        decoded_results.append({
            "databits": databits,
            "address_bits": address_bits,
            "flipped_reversed_address": reversed_address,
            "decimal_address": decimal_address,
            "count": count
        })

    # Sort results by decimal_address
    decoded_results_sorted = sorted(decoded_results, key=lambda x: x["decimal_address"])

    print("\nDecoded Quantum Results (Sorted by Address):")
    for result in decoded_results:
        print(f"Address (Decimal): {result['decimal_address']}, Data bits: {result['databits']}, "
              f"Flipped & Reversed: {result['flipped_reversed_address']}, Count: {result['count']}")


if __name__ == "__main__":
    main()


